/*
 *	htanaly.cc
 */

/*	Copyright (C) 1999, 2000, 2001 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "analy.h"
#include "analy_names.h"
#include "global.h"
#include "htanaly.h"
#include "htapp.h"
#include "htdialog.h"
#include "hthist.h"
#include "htidle.h"
#include "htiobox.h"
#include "htkeyb.h"
#include "htstring.h"
#include "httag.h"
#include "httree.h"
#include "tools.h"

#include <stdlib.h>
#include <string.h>

// shouldnt be here
#include "alphadis.h"
#include "x86dis.h"

//test
#include "out_html.h"
#include "store.h"
#include "htsearch.h"


/*
 *	symbolbox
 */

void symbolbox::init(bounds *b, analyser *Analy)
{
	analy = Analy;
	ht_listbox::init(b);
	str = (char *)smalloc(1024); // Buffer overflow...
}

void symbolbox::done()
{
	ht_listbox::done();
}

int  symbolbox::calc_count()
{
	return analy->getlabelcount();
}

int  symbolbox::cursor_adjust()
{
	return 18;
}

void *symbolbox::getfirst()
{
	return analy->enum_labels(NULL);
}

void *symbolbox::getlast()
{
	return analy->enum_labels_back(NULL);
}

void *symbolbox::getnext(void *entry)
{
	if (!entry) return NULL;
	return analy->enum_labels(((tlabel *)entry)->name);
}

void *symbolbox::getnth(int n)
{
	// slow!!!!!!!!!!
	// n==1 =^= 1th entry
	tlabel *e = NULL;
	for (int i=0; i<n; i++) {
		e = analy->enum_labels(((tlabel *)e)->name);
		if (!e) break;
	}
	return e;
}

void *symbolbox::getprev(void *entry)
{
	if (!entry) return NULL;
	return analy->enum_labels_back(((tlabel *)entry)->name);
}

char *symbolbox::getstr(int col, void *entry)
{
	if (!entry) return NULL;
	tlabel *l = ((tlabel *)entry);
	switch (col) {
		case 0:
			sprintf(str, "%8x", l->addr->addr);
			break;
		case 1:
			strcpy(str, label_type_short(l->type));
			break;
		case 2:
			strcpy(str, l->name);
			break;
	}
	return str;
}

bool symbolbox::idle()
{
	if ((idle_count % 500)==0) {
		update();
		redraw();
		idle_count = 1;
		return 1;
	}
	idle_count++;
	return 0;
}

int	symbolbox::num_cols()
{
	return 3;
}

void *symbolbox::quickfind(char *s)
{
	tlabel *tmp = analy->enum_labels(s);
	if (tmp) {
		int slen = strlen(s);
		int tlen = strlen(tmp->name);
		if (slen > tlen) return NULL;
		if (strncmp(tmp->name, s, slen)==0) return tmp;
		return NULL;
	} else {
		return NULL;
	}
}


char	*symbolbox::quickfind_completition(char *s)
{
	if (analy->findlabel(s)) {
		return ht_strdup(s);
	}
	tlabel *tmp = analy->enum_labels(s);
	if (!tmp) {
		return ht_strdup(tmp->name);
	}
	tlabel *tmp2 = analy->enum_labels(tmp->name);
	if (!tmp2) {
		return ht_strdup(tmp->name);
	}
	int slen = strlen(s);
	if (!ht_strncmp(tmp->name, tmp2->name, slen)==0) {
		return ht_strdup(tmp->name);
	}
	char *res = (char *)smalloc(1024); // buffer bla
	strcpy(res, tmp->name);
	while (tmp2 && (ht_strncmp(tmp2->name, s, slen)==0)) {
//		fprintf(stdout, "while(%s, %s, %d)\n", tmp2->name, s, slen);
		int a = strccomm(res, tmp2->name);
		res[a] = 0;
		tmp2 = analy->enum_labels(tmp2->name);
	}
	return res;
}


/*
 *	CLASS call_chain
 */
void	call_chain::init(bounds *b, analyser *Analy, ADDR a, char *desc)
{
	ht_treeview::init(b, desc);
	VIEW_DEBUG_NAME("call_chain");
	analy = Analy;
	root = create_node(a);
}

void call_chain_done(call_chain_node *n)
{
	while (n) {
		call_chain_done(n->child);
		call_chain_node *temp = n->next;
		free(n);
		n = temp;
	}
}

void	call_chain::done()
{
	call_chain_done(root);
	ht_treeview::done();
}

void	call_chain::adjust(void *node, bool expand)
{
	((call_chain_node*)node)->expanded = expand;
}

call_chain_node *call_chain::create_node(ADDR A)
{
	call_chain_node *n = (call_chain_node *)smalloc(sizeof(call_chain_node));
	n->next = NULL;
	n->prev = NULL;
	n->child = NULL;
	n->examined = false;
	n->xa = A;
	n->faddr = analy->findaddrfunc(A);
	assert(n->faddr);
	n->fa = n->faddr->addr;
	n->expanded = false;
	return n;
}

void call_chain::examine_node(call_chain_node *n)
{
	n->examined = true;
	if (has_children(n)) {
		txref *x = n->faddr->xreflist;
		call_chain_node *nn = n->child = create_node(x->addr);
		x = x->next;
		while (x) {
			nn->next = create_node(x->addr);
			x = x->next;
			nn = nn->next;
		}
	}
}

void	*call_chain::get_child(void *node, int i)
{
	call_chain_node *p;
	if (node) {
		if (!((call_chain_node *)node)->examined) examine_node((call_chain_node*)node);
		p = ((call_chain_node *)node)->child;
	} else {
		p = root;
	}
	while (p && (--i)) p = p->next;
	return p;
}

void	*call_chain::get_next_node(void *node)
{
	return ((call_chain_node*)node)->next;
}

void	*call_chain::get_prev_node(void *node)
{
	return ((call_chain_node*)node)->prev;
}

void	*call_chain::get_root()
{
	return root;
}

char	*call_chain::get_text(void *node)
{
	static char stupid[1024]; // FIXME: static var + buffer bla
	call_chain_node *n=(call_chain_node*)node;
	sprintf(stupid, "%s+%x (0x%x)", n->faddr->label?n->faddr->label->name:"unknown", n->xa-n->faddr->addr, n->xa);
	return stupid;
}

bool	call_chain::has_children(void *node)
{
	return ((call_chain_node*)node)->faddr->xreflist != NULL;
}

bool	call_chain::is_expanded(void *node)
{
	return ((call_chain_node*)node)->expanded;
}

void	call_chain::select_node(void *node)
{
}


/////////////////////////////////////////////////////////////////////////////

/*
 *
 */
void analy_infoline::init(bounds *b, ht_aviewer *A, char *Format)
{
	ht_statictext::init(b, 0, align_left);
	VIEW_DEBUG_NAME("analy_infoline");
	analy = A;
	displayformat = ht_strdup(Format);
	s = (char *)smalloc(1024); // FIXME: buffer bla
}

void analy_infoline::done()
{
	free(displayformat);
	free(s);
	ht_statictext::done();
}

char *analy_infoline::gettext()
{
	if (valid()) {
		char *sec = analy->analy->getaddrsectionname(addr);
		if (fofs!=INVALID_FILE_OFS) {
			taddr *a = analy->analy->findaddrfunc(addr);
			char *func = (a) ? ((a->label) ? a->label->name : NULL): NULL;
			
//			sprintf(s, displayformat, (sec) ? sec : "", fofs,  analy->analy->getdisasmstr(addr), (func)?func:"");
			char *d = displayformat;
			char *ss = s;
			while (*d) {
				if (*d=='%') {
					d++;
					switch (*d) {
						case ANALY_STATUS_ARG_SECTION:
							if (sec) ss+=sprintf(ss, "%s", sec);
							break;
						case ANALY_STATUS_ARG_FILEOFFSET:
							ss+=sprintf(ss, "%08x", fofs);
							break;
						case ANALY_STATUS_ARG_RAW_UNASM:
							ss+=sprintf(ss, "%s", analy->analy->getdisasmstr(addr));
							break;
						case ANALY_STATUS_ARG_FUNCTION:
							if (func) ss+=sprintf(ss, "%s+%x", func, addr-a->addr);
							break;
						case ANALY_STATUS_ARG_OFFSET:
							ss+=sprintf(ss, "%08x", addr);
							break;
						case '%':
							*ss++ = '%';
							break;
						default:
							ss += sprintf(ss, " error in format ");
					}
				} else {
					*ss++ = *d;
				}
				d++;
			}
			*ss = 0;
		} else {
			if (!sec) {
				strcpy(s, "[not in file]");
			} else {
				sprintf(s, "<%s> [not in file]", sec);
			}
		}
		return s;
	} else {
		return "<no analyser>";
	}
}

void analy_infoline::update(ADDR cursor_addr, ADDR ecursor_addr)
{
	if (valid()) {
		if (ecursor_addr) {
			fofs = ecursor_addr;
		} else {
			fofs = analy->analy->fileaddr(cursor_addr);
		}
		addr = cursor_addr;
	} else {
		fofs = INVALID_FILE_OFS;
		addr = INVALID_ADDR;
	}
	dirtyview();
	redraw();
}

bool analy_infoline::valid()
{
	return ((analy)&&(analy->analy));
}

void ht_aviewer::init(bounds *b, char *desc, int caps, ht_streamfile *file, ht_format_group *format_group, analyser *Analy)
{
	analy=Analy;
	analy->setdisplaymode(ANALY_TRANSLATE_SYMBOLS+ANALY_COLLAPSE_XREFS, 0);
	ht_uformat_viewer::init(b, desc, caps, file, format_group);
	infoline=NULL;
	idle_count=0;
	last_active=1;
	register_idle_object(this);
}

void ht_aviewer::done()
{
	unregister_idle_object(this);
	ht_uformat_viewer::done();
	analy->done();
	delete analy;
}

void ht_aviewer::attach_infoline(analy_infoline *V)
{
	infoline = V;
}

bool ht_aviewer::address_to_offset(fmt_vaddress addr, FILEOFS *ofs)
{
	FILEOFS o=analy->fileaddr(addr);
	if (o!=INVALID_FILE_OFS) {
		*ofs=o;
		return 1;
	}
	return 0;
}

char *ht_aviewer::func(UINT i, bool execute)
{
	switch (i) {
		case 6: {
			if (execute) {
				bounds b;
				BOUNDS_ASSIGN(b, 10, 5, 60, 15);
				ht_dialog *dialog = new ht_dialog();
				dialog->init(&b, "symbols", FS_KILLER | FS_TITLE | FS_MOVE);
				/* pull down */
				BOUNDS_ASSIGN(b, 30, 0, 20, 1);
				ht_listpopup *lp = new ht_listpopup();
				lp->init(&b);
				lp->insertstring("show all");
				lp->insertstring("only functions");
				lp->insertstring("only labels");
				lp->insertstring("only offsets (code)");
				lp->insertstring("only data");
//				dialog->insert(lp);
				/* text */
				BOUNDS_ASSIGN(b, 1, 0, 30, 1);
				ht_statictext *text = new ht_statictext();
				text->init(&b, " Address   Type   Name", 0);
				dialog->insert(text);
				/* list */
				BOUNDS_ASSIGN(b, 1, 1, 56, 12);
				symbolbox *sym=new symbolbox();
				sym->init(&b, analy);
				dialog->insert(sym);
				register_idle_object(sym);
				int r = dialog->run(0);
				unregister_idle_object(sym);
				if (r) {
					// goto selected symbol
					ht_listbox_data d;
					sym->databuf_get(&d);
					fmt_vaddress vaddr;
					if (string_to_address(((tlabel *)(sym->getbyid(d.cursor_id)))->name, &vaddr)) {
						goto_address(vaddr);
					}
				}
				dialog->done();
				delete dialog;
			}
			return "symbols";
		}
		default:
			return ht_uformat_viewer::func(i, execute);
	}
	return 0;
}

int ht_aviewer_symbol_to_addr(ht_aviewer *aviewer, char **s, dword *v)
{
	if (**s == '@') {
		(*s)++;
		if (bnstr(s, v, 10)) {
			if (!aviewer->offset_to_address(*v, v)) {
				return 0;
			}
			return 1;
		}
		// invalid number after @
		return 0;
	} else {
		char *k=ht_strdup(*s);
		char *t=k;
		while (!strchr("+-*/ \t[]", *t) && *t) t++;
		char temp=*t;
		*t=0;
		if ((*k == '$') && (k[1] == 0)) {
			*t=temp;
			*s+= t-k;
			aviewer->get_current_address(v);
			free(k);
			return 1;
		}
		tlabel *l = aviewer->analy->findlabel(k);
		*t=temp;
		if (l) {
			// Label
			*s+=t-k;
			*v=l->addr->addr;
			free(k);
			return 1;
/*		} else if (**s=='.') {
			// RVA
			if (bnstr(s, v, 16)) {
				return 1;
			}*/
		}
		free(k);
	}
	return 0;
}

void ht_aviewer::handlemsg(htmsg *msg)
{
	char str[1024];
	switch (msg->msg) {
		case msg_keypressed:
			switch (msg->data1.integer) {
				case K_Control_A:
					sendmsg(cmd_analy_call_assembler);
					clearmsg(msg);
					return;
				case K_Control_B:
					clearmsg(msg);
					return;
				case K_Control_D:
					sendmsg(cmd_analy_del_addr_bindings);
					clearmsg(msg);
					return;
				case K_Control_F:
					sendmsg(cmd_analy_this_function);
					clearmsg(msg);
					return;
				case K_Control_L:
					sendmsg(cmd_analy_previous_label);
					clearmsg(msg);
					return;
				case K_Control_O: {
					ht_file *s = new ht_file();
					s->init("__test.htm", FAM_CREATE+FAM_WRITE);
//					generate_html_output(analy, s, 0x406400, 0x406400+0x1000);
					generate_html_output(analy, s, 0x10020c00, 0x10020c9e);
					s->done();
					delete s;
					clearmsg(msg);
					return;
				}
				case K_Control_S: {
					ht_file *s = new ht_file();
					s->init("test.sav", /*FAM_READ);*/FAM_CREATE+FAM_WRITE);
					ht_object_stream_bin *o = new ht_object_stream_bin();
					o->init(s, true);
//                         app->store(o);
					PUT_OBJECT(o, analy);
/*                         o->put_info("test1");
					o->put_int_dec(1, 4, "testa");
					o->put_int_hex(10, 4, "testb");
					s->seek(0);*/
//					int a = o->get_int_dec(4, "testa");
/*					int b = o->get_int_dec(4, "testb");
					int c = o->get_error();
					infobox("%d, %d -> %d", a, b, c);*/
					o->done();
					delete o;
					s->done();
					delete s;
					clearmsg(msg);
					return;
				}
				case K_Control_T:
					sendmsg(cmd_analy_call_chain);
					clearmsg(msg);
					return;
				case 'c':
					if (!edit()) {
						sendmsg(cmd_analy_continue);
						clearmsg(msg);
						return;
					}
					break;
				case '#':
					sendmsg(cmd_analy_comments);
					clearmsg(msg);
					return;
				case 'n':
					sendmsg(cmd_analy_name_addr);
					clearmsg(msg);
					return;
				case 'x':
					sendmsg(cmd_analy_xrefs);
					clearmsg(msg);
					return;
				case 'f':
					if (!edit()) {
						sendmsg(cmd_analy_follow);
						clearmsg(msg);
						return;
					}
					break;
				case 'F':
					if (!edit()) {
						sendmsg(cmd_analy_follow_ex);
						clearmsg(msg);
						return;
					}
					break;
				case K_Delete:
					sendmsg(cmd_analy_del_addr_bindings);
					clearmsg(msg);
					return;
			}
			break;
		case cmd_analy_call_assembler: {
			assembler *a = analy->create_assembler();
			if (!a) {
				// select assembler
				infobox("no assembler available.");
				clearmsg(msg);
				return;
			}
			fmt_vaddress current_address;
			get_current_address(&current_address);

			a->set_imm_eval_proc((int(*)(void *context, char **s, dword *v))ht_aviewer_symbol_to_addr, (void*)this);
			
			char instr[257] = "";
			strcpy(instr, analy->getdisasmstr_formatted(current_address));
			asm_insn *insn = a->alloc_insn();
			asm_code *ac = NULL;
			while (inputbox(a->get_name(), "instruction:", instr, 255, HISTATOM_ASSEMBLER)) {
				if ((a->prepare_str(insn, instr)) && (ac = a->encode(insn, 0, analy->mapaddr(current_address)))) {
					break;
				} else {
					errorbox("%s: %s", a->get_name(), a->get_error_msg());
				}
			}
			if (ac) {
				bool ok=true;
				asm_code *chosen_ac = ac;
				if (ac->next) {
					// choose from list if ambigous
					bounds b;
					BOUNDS_ASSIGN(b, 10, 5, 60, 15);
					ht_dialog *dialog = new ht_dialog();
					dialog->init(&b, "choose opcode", FS_KILLER | FS_TITLE | FS_MOVE);
					BOUNDS_ASSIGN(b, 1, 0, 56, 1);
					ht_statictext *text = new ht_statictext();
					text->init(&b, " opcode (& disassembly)", 0);
					dialog->insert(text);
					BOUNDS_ASSIGN(b, 1, 1, 56, 12);
					ht_text_listbox *list=new ht_text_listbox();
					list->init(&b, 2, 0);
					asm_code *ac2 = ac;
					UINT aci=0;
					while (ac2) {
						char s[1024]="", *tmp = s;
						for (UINT i=0; i<ac2->size; i++) {
							tmp += sprintf(tmp, "%02x ", ac2->data[i]);
						}
						if (analy->disasm) {
							OPCODE *o=analy->disasm->decode((byte *)ac2->data, ac2->size, analy->mapaddr(current_address));
							tmp = analy->disasm->strf(o, X86DIS_STYLE_HEX_NOZEROPAD+X86DIS_STYLE_HEX_ASMSTYLE, DISASM_STRF_SMALL_FORMAT);
						} else {
							tmp = "<no disassembler>";
						}
						list->insert_str(aci, s, tmp);
						ac2 = ac2->next;
						aci++;
					}
					list->update();
					dialog->insert(list);
					int r = dialog->run(0);
					ok = r;
					if (r) {
						ht_listbox_data d;
						list->databuf_get(&d);
						ht_text_listbox_item *i=(ht_text_listbox_item *)list->getbyid(d.cursor_id);
						asm_code *ac3 = ac;
						int ac3i=0;
						while (ac3) {
							if (ac3i==i->id) {
								chosen_ac=ac3;
								break;
							}
							ac3=ac3->next;
							ac3i++;
						}
					}
					dialog->done();
					delete dialog;
				}
				if (ok) {
					baseview->sendmsg(cmd_edit_mode_i, file, NULL);
					if (edit()) {
						vwrite(current_address, chosen_ac->data, chosen_ac->size);
					}
				}
			}
			free(insn);
			if (a) {
				a->done();
				delete a;
			}
			clearmsg(msg);
			return;
		}
		case cmd_analy_this_function: {
			taddr *a = analy->findaddrfunc(cursor_id1);
			if (a) {
				goto_address(a->addr, this);
			} else {
				errorbox("Address %08lx doesn't belong to a function.", cursor_id1);
			}
			clearmsg(msg);
			return;
		}
		case cmd_analy_previous_label: {
			taddr *a = analy->findaddrlabel(cursor_id1);
			if (a) {
				goto_address(a->addr, this);
			} else {
				errorbox("Address %08lx doesn't belong to a label.", cursor_id1);
			}
			clearmsg(msg);
			return;
		}
		case cmd_analy_continue:
			analy->continueanalysisat(cursor_id1);
			clearmsg(msg);
			return;
		case cmd_analy_comments: {
			fmt_vaddress current_address;
			get_current_address(&current_address);
			tcomment *c = analy->getcomments(current_address);
			if (c) {
				tcomment *d = c;
				int size = 3;
				while (d) {
					size += strlen(d->text)+1;
					d = d->next;
				}
				char *text = (char *)smalloc(size);
				char *t = text;
				t += sprintf(t, "\el");
				d = c;
				while (d) {
					t += sprintf(t, "%s\n", d->text);
					d = d->next;
				}
				*t = 0;
				infobox_c(text);
				free(text);
			} else {
				errorbox("no comments.");
			}
			clearmsg(msg);
			return;
		}
		case cmd_analy_name_addr: {
			fmt_vaddress current_address;
			get_current_address(&current_address);
			dword addr=(dword)current_address;
			if (analy->validaddr(addr, scvalid)) {
				char n[255];
				tlabel *l = analy->getaddrlabel(addr);
				if (l) strcpy(n, l->name); else n[0] = 0;
				sprintf(str, "name for address %08x", addr);
				if (inputbox(str, "label name:", n, 255, HISTATOM_NAME_ADDR)) {
					if (n[0]) {
						if (valid_name(n)) {
							char *n2 = ht_strdup(n);
							if (!analy->assignlabel(addr, n2, (l)?l->type:label_loc)) {
								l = analy->findlabel(n);
								errorbox("Label `%s' already exists at address %08x!", n, l->addr->addr);
								free(n2);
							}
						} else {
							errorbox("`%s' is an invalid label name.", n);
						}
					} else {
						if (l) {
							// delete label if applicable
							if (confirmbox("Really delete label `%s' at address %08x?", l->name, addr)==button_yes) {
								analy->deletelabel(addr);
							}
						}
					}
				}
			} else {
				errorbox("Address %08lx is invalid!", addr);
			}
			clearmsg(msg);
			return;
		}
		case cmd_analy_xrefs:
			show_xrefs(cursor_id1);
			clearmsg(msg);
			return;
		case cmd_analy_follow: {
			ADDR *ptr = (ADDR *)analy->bufptr(cursor_id1);
			ADDR p = *ptr;
			if (analy->validaddr(p, scvalid)) {
				goto_address(p, this);
			} else {
				errorbox("Follow: address %08lx is invalid!", p);
			}
			clearmsg(msg);
			return;
		}
		case cmd_analy_follow_ex:
			clearmsg(msg);
			return;
		case cmd_analy_pause_resume:
			clearmsg(msg);
			return;
		case cmd_analy_del_addr_bindings: {
			fmt_vaddress current_address;
			get_current_address(&current_address);
			if (confirmbox("Forget about address' %08x bindings?\n(name, comments, xrefs, etc.)", current_address)==button_yes) {
				analy->deleteaddr((dword)current_address);
			}
			clearmsg(msg);
			return;
		}
		case cmd_edit_mode:
		case cmd_view_mode:
			if (edit()) {
				analy->setdisplaymode(ANALY_EDIT_BYTES, 0);
			} else {
				analy->setdisplaymode(0, ANALY_EDIT_BYTES);
			}
			ht_uformat_viewer::handlemsg(msg);
			return;
		case cmd_analy_call_chain: {
			fmt_vaddress current_address;
			get_current_address(&current_address);
			show_call_chain((ADDR) current_address);
			return;
		}
	}
	ht_uformat_viewer::handlemsg(msg);
	switch (msg->msg) {
		case msg_draw:
			if (infoline) {
				FILEADDR a;
				fmt_vaddress b;
				infoline->update((get_current_address(&b)) ? b : 0, (get_current_offset(&a)) ? a : 0);
			}
			break;
	}
}

bool ht_aviewer::idle()
{
	last_active=analy->active;
	analy->continueanalysis();
	if (last_active && !analy->active) {
		LOG("%s: analyser finished after %d ops.", analy->getname(), analy->ops_parsed);
		dirtyview();
		app->sendmsg(msg_draw, 0);
	}
	if (analy->active) {
		idle_count++;
		if (idle_count%50==0) {
			dirtyview();
			app->sendmsg(msg_draw, 0);
		}
	}
	return analy->active;
}

bool ht_aviewer::offset_to_address(FILEOFS ofs, fmt_vaddress *addr)
{
	*addr = analy->vaddr(ofs);
	return (*addr != INVALID_ADDR);
}

int ht_aviewer::ref_sel(ID id_low, ID id_high)
{
	switch (id_high) {
		case 0:
			return goto_address(id_low, this);
		case 1:
			show_xrefs((ADDR) id_low);
			return 0;
	}
	return 0;
}

void ht_aviewer::show_call_chain(ADDR Addr)
{
	taddr *a = analy->findaddrfunc(Addr);
	if (!a) return;
//	Addr = a->addr;

	bounds b;
	BOUNDS_ASSIGN(b, 10, 5, 60, 15);
	char str[100];
	sprintf(str, "call chain of address %08x", Addr);
	ht_dialog *dialog = new ht_dialog();
	dialog->init(&b, str, FS_KILLER | FS_TITLE | FS_MOVE);
	BOUNDS_ASSIGN(b, 1, 0, 56, 10);
	ht_statictext *text = new ht_statictext();
	sprintf(str, (a->label)?"function %s %s":"address %08x %s", (a->label)?(int)a->label->name:(int)a->addr, "is referenced by ...");
	text->init(&b, str, 0);
	dialog->insert(text);
	BOUNDS_ASSIGN(b, 1, 1, 56, 10);
	call_chain *cc;
	NEW_OBJECT(cc, call_chain, &b, analy, Addr, NULL);
	cc->adjust(cc->get_root(), true);
	dialog->insert(cc);
	int r = dialog->run(0);
	if (r) {
	}
}

void ht_aviewer::show_xrefs(ADDR Addr)
{
	txref *x = analy->getxrefs(Addr);
	if (x) {
		bounds b;
		BOUNDS_ASSIGN(b, 10, 5, 60, 15);
		char str[100];
		sprintf(str, "xrefs of address %08x", Addr);
		ht_dialog *dialog = new ht_dialog();
		dialog->init(&b, str, FS_KILLER | FS_TITLE | FS_MOVE);
		BOUNDS_ASSIGN(b, 1, 0, 56, 10);
		ht_statictext *text = new ht_statictext();
		text->init(&b, " xref to   type     from function", 0);
		dialog->insert(text);
		BOUNDS_ASSIGN(b, 1, 1, 56, 10);
		ht_text_listbox *list;
		NEW_OBJECT(list, ht_text_listbox, &b, 3, 2);
		char str2[1024];
		while (x) {
			sprintf(str, "%8x", x->addr);
			taddr *a = analy->findaddrfunc(x->addr);
			char *func = (a) ? ((a->label) ? a->label->name : NULL): NULL;
			if (func) sprintf(str2, "%s+%x", func, x->addr-a->addr); else sprintf(str2, "?");
			list->insert_str(x->addr, str, xref_type(x->type), str2);
			x = x->next;
		}
		list->update();
		dialog->insert(list);
		int r = dialog->run(0);
		if (r) {
			ht_listbox_data data;
			list->databuf_get(&data);
			goto_address(data.cursor_id, this);
		}
		dialog->done();
		delete dialog;
	} else {
		if (confirmbox("No xrefs for address %08lx!\nSearch for xrefs?", Addr)==button_yes) {
			char str[100];
			sprintf(str, "%xh", Addr);
			ht_regex_search_request *q=new ht_regex_search_request(SC_VISUAL, SF_REGEX_CASEINSENSITIVE, str);
			ht_visual_search_result *r=(ht_visual_search_result *)vsearch(q, 0, 0xffffffff);
			while (r) {
				analy->addxref(Addr, r->address, xrefoffset);
				delete r;
				fmt_vaddress na;
				next_logical_address(r->address, &na);
				r = (ht_visual_search_result *)vsearch(q, na, 0xffffffff);
			}
			delete q;
		}
	}
}

bool ht_aviewer::string_to_address(char *string, fmt_vaddress *vaddr)
{
	ADDR a;
	int ev=eval_int_expr_custom(string, &a, this, (int (*)(void *context, char **s, dword *v))ht_aviewer_symbol_to_addr);
	if (ev) {
		if (analy->validaddr(a, scvalid)) {
			*vaddr=a;
			return 1;
		} else {
			sprintf(globalerror, "address %08x is invalid", a);
		}
	} else {
		strcpy(globalerror, eval_int_errstr);
	}
	return 0;
}

/*
 *	CLASS ht_analy_sub
 */

void ht_analy_sub::init(ht_streamfile *file, analyser *A, ADDR Lowestaddress, ADDR Highestaddress)
{
	ht_sub::init(file);
	analy = A;
	lowestaddress=Lowestaddress;
	highestaddress=Highestaddress-1;
}

void ht_analy_sub::done()
{
	ht_sub::done();
}

bool	ht_analy_sub::convert_addr_to_id(fmt_vaddress addr, ID *id1, ID *id2)
{
	ADDR a=addr; /* FIXME: */
	if (analy->validaddr(a, scvalid)) {
		*id1=addr;
		*id2=0;
		return true;
	}
	return false;
}

bool	ht_analy_sub::convert_id_to_addr(ID id1, ID id2, fmt_vaddress *addr)
{
	*addr=id1;
	return 1;
}

void	ht_analy_sub::first_line_id(ID *id1, ID *id2)
{
	*id1=lowestaddress;
	*id2=0;
}

bool ht_analy_sub::getline(char *line, ID id1, ID id2)
{
	int len;
	analy->getformattedline(id1, id2+1, line, &len);
	return 1;
}

void	ht_analy_sub::last_line_id(ID *id1, ID *id2)
{
	*id1=highestaddress;
	*id2=0;
}

int	ht_analy_sub::next_line_id(ID *id1, ID *id2, int n)
{
	int len;
	int res=0;
	while (n--) {
		if (analy->getformattedline_length(*id1, *id2+2, &len)) {
			(*id2)++;
		} else {
			analy->getformattedline_length(*id1, *id2+1, &len);
			if (*id1>=highestaddress) return res;
			*id1 += len;
			*id2 = 0;
		}
		res++;
	}
	return res;
}

int	ht_analy_sub::prev_line_id(ID *id1, ID *id2, int n)
{
/* FIXME: interesting solution! */
	ADDR addrbuf[1024];
	int	linebuf[1024];
	int	i=0;
	int len;

/*
 *	24==1*2^3*3
 *   so it "supports" opcodes of length 1, 2, 3, 4, 6, 8 and 12
 *   (and of course 24)
 */
	ADDR	l=n*24;

/*
 *	120==24*5
 *   so 120 additionally "supports" opcodes of length 5, 10 and 15
 *   (and some more)
 */
	if (l<120) l=120;

	/* look backwards */
//     fprintf(stdout, "============= up ==============================\n");
//     fprintf(stdout, "%d x from %lx line %ld\n", n, *id1, *id2);
	ADDR tmpaddr = (*id1) - l;
	ADDR tmpaddr2;
	dword tmpline;
	first_line_id(&tmpaddr2, &tmpline);
	if ((l > (*id1)) || (tmpaddr < tmpaddr2)) tmpaddr = tmpaddr2;
	ADDR	nextaddr = tmpaddr;
	while (1) {
		if (tmpaddr >= (*id1)) {
			if ((tmpline >= (*id2)) || (tmpaddr > *id1)) break;
		}
		if (analy->getformattedline_length(tmpaddr, tmpline+1, &len)) {
			nextaddr += len;
			tmpline++;
		} else {
			tmpaddr = nextaddr;
			tmpline = 0;
			continue;
		}
//          fprintf(stdout, "%2d. %lx line %ld\n", i, tmpaddr, tmpline-1);
		addrbuf[i&(1023)] = tmpaddr;
		linebuf[i&(1023)] = tmpline-1;
		i++;
	}

	if (!i) return 0;

	if (i >= n) {
//     	fprintf(stdout, "result1: %lx line %d (%d)\n", addrbuf[i-n], linebuf[i-n], n);
		(*id1) = addrbuf[(i-n)&(1023)];
		(*id2) = linebuf[(i-n)&(1023)];
		return n;
	} else {
//     	fprintf(stdout, "result2: %lx line %d (%d)\n", addrbuf[0], linebuf[0], i);
		(*id1) = addrbuf[0];
		(*id2) = linebuf[0];
		return i;
	}
}

/*int ht_analy_sub::search(ht_search *s, STREAMOFS start, STREAMOFS end, STREAMOFS *offset, dword *size)
{
	return SR_NOT_SUPPORTED;
}*/

