/*
 *	HT Editor
 *	stream.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htdebug.h"
#include "stream.h"
#include "tools.h"

#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/*
 *	CLASS ht_stream
 */

void	ht_stream::init()
{
	   stream_error_func=0;
	   error=0;
	   access_mode=0;
}

void	ht_stream::done()
{
}

int	ht_stream::call_error_func()
{
	if (stream_error_func) return stream_error_func(this); else
		return SERR_FAIL;
}

UINT	ht_stream::get_access_mode()
{
	return access_mode;
}

int	ht_stream::get_error()
{
	return error;
}

char	*ht_stream::get_name()
{
	return NULL;
}

int	ht_stream::set_access_mode(UINT a)
{
	access_mode=a;
	return 1;
}

void	ht_stream::set_error(int e)
{
	error=e;
}

void	ht_stream::set_error_func(stream_error_func_ptr s)
{
	   stream_error_func=s;
}

UINT	ht_stream::read(void *buf, UINT size)
{
	return 0;
}

UINT	ht_stream::write(void *buf, UINT size)
{
	return 0;
}

/*
 *	ht_layer_stream
 */

void	ht_layer_stream::init(ht_stream *s, bool own)
{
	ht_stream::init();
	stream=s;
	own_stream=own;
}

void	ht_layer_stream::done()
{
	if (own_stream) {
		stream->done();
		delete stream;
	}
	ht_stream::done();
}

UINT	ht_layer_stream::get_access_mode()
{
	return stream->get_access_mode();
}

int	ht_layer_stream::get_error()
{
	return stream->get_error();
}

char	*ht_layer_stream::get_name()
{
	return stream->get_name();
}

int	ht_layer_stream::set_access_mode(UINT access_mode)
{
	return stream->set_access_mode(access_mode);
}

void	ht_layer_stream::set_error(int error)
{
	stream->set_error(error);
}

void	ht_layer_stream::set_error_func(stream_error_func_ptr stream_error_func)
{
	stream->set_error_func(stream_error_func);
}

UINT	ht_layer_stream::read(void *buf, UINT size)
{
	return stream->read(buf, size);
}

UINT	ht_layer_stream::write(void *buf, UINT size)
{
	return stream->write(buf, size);
}

/*
 *	CLASS ht_object_stream
 */
 
void ht_object_stream::init(ht_stream *s)
{
	ht_layer_stream::init(s, false);
}

void ht_object_stream::done()
{
	ht_layer_stream::done();
}

int  ht_object_stream::get_int(int size, char *desc)
{
	return get_int_hex(size, desc);
}

UINT ht_object_stream::record_start(UINT size)
{
	return 0;
}

void ht_object_stream::record_end(UINT a)
{
}

void ht_object_stream::put_int(int a, int size, char *desc)
{
	put_int_hex(a, size, desc);
}

/*
 *	CLASS ht_streamfile
 */

int	ht_streamfile::cntl(UINT cmd, ...)
{
	va_list vargs;
	   va_start(vargs, cmd);
	int ret=vcntl(cmd, vargs);
	   va_end(vargs);
	   return ret;
}

int	ht_streamfile::extend(UINT newsize)
{
	return ENOSYS;
}

UINT	ht_streamfile::get_size()
{
	return 0;
}

void	ht_streamfile::pstat(pstat_t *s)
{
	s->caps=0;
}

int	ht_streamfile::seek(FILEOFS offset)
{
	return ENOSYS;
}

FILEOFS	ht_streamfile::tell()
{
	return 0;
}

int	ht_streamfile::truncate(UINT newsize)
{
	return ENOSYS;
}

int	ht_streamfile::vcntl(UINT cmd, va_list vargs)
{
	return ENOSYS;
}

/*
 *	CLASS ht_layer_streamfile
 */

void	ht_layer_streamfile::init(ht_streamfile *s, bool own)
{
	ht_streamfile::init();
	streamfile=s;
	own_streamfile=own;
}

void	ht_layer_streamfile::done()
{
	if (own_streamfile) {
		streamfile->done();
		delete streamfile;
	}
	ht_streamfile::done();
}

UINT	ht_layer_streamfile::get_access_mode()
{
	return streamfile->get_access_mode();
}

int	ht_layer_streamfile::get_error()
{
	return streamfile->get_error();
}

char	*ht_layer_streamfile::get_name()
{
	return streamfile->get_name();
}

UINT	ht_layer_streamfile::get_size()
{
	return streamfile->get_size();
}

void	ht_layer_streamfile::pstat(pstat_t *s)
{
	streamfile->pstat(s);
}

UINT	ht_layer_streamfile::read(void *buf, UINT size)
{
	return streamfile->read(buf, size);
}

int	ht_layer_streamfile::seek(FILEOFS offset)
{
	return streamfile->seek(offset);
}

int	ht_layer_streamfile::set_access_mode(UINT access_mode)
{
	return streamfile->set_access_mode(access_mode);
}

void	ht_layer_streamfile::set_error(int error)
{
	streamfile->set_error(error);
}

void	ht_layer_streamfile::set_error_func(stream_error_func_ptr stream_error_func)
{
	streamfile->set_error_func(stream_error_func);
}

FILEOFS	ht_layer_streamfile::tell()
{
	return streamfile->tell();
}

int	ht_layer_streamfile::truncate(UINT newsize)
{
	return streamfile->truncate(newsize);
}

int	ht_layer_streamfile::vcntl(UINT cmd, va_list vargs)
{
	return streamfile->vcntl(cmd, vargs);
}

UINT	ht_layer_streamfile::write(void *buf, UINT size)
{
	return streamfile->write(buf, size);
}

/*
 *	CLASS ht_file
 */

void	ht_file::init(char *fn, UINT am)
{
	ht_streamfile::init();
	filename=strdup(fn);
	offset=0;
	file=0;
	access_mode=0;
	set_access_mode(am);
}

void	ht_file::done()
{
	if (filename) free(filename);
	if (file) fclose(file);
	ht_streamfile::done();
}

int	ht_file::extend(UINT newsize)
{
	return ENOSYS;
}

char	*ht_file::get_name()
{
	return filename;
}

UINT	ht_file::get_size()
{
	int t=tell();
	fseek(file, 0, SEEK_END);	/* zero is allowed ;-) */
	int r=ftell(file);
	fseek(file, t, SEEK_SET);
	return r;
}

void	ht_file::pstat(pstat_t *s)
{
	sys_pstat(s, filename);
}

UINT	ht_file::read(void *buf, UINT size)
{
	UINT r=fread(buf, 1, size, file);
	offset+=r;
	return r;
}

int	ht_file::seek(FILEOFS o)
{
	if (o==offset) return 0;
	int r=fseek(file, o, SEEK_SET);
	if (!r) offset=o;
	return r;
}

int	ht_file::set_access_mode(UINT am)
{
	UINT orig_access_mode=access_mode;
	int r=set_access_mode_internal(am);
	if (!r) set_access_mode_internal(orig_access_mode);
	return r;
}

int	ht_file::set_access_mode_internal(UINT am)
{
RETRY:
	if (access_mode==am) return 1;
	if (file) {
		fclose(file);
		file=0;
	}
	access_mode=FAM_UNDEFINED;
	char *mode=0;
	if (am & FAM_CREATE) {
		if (am & FAM_WRITE) mode="wb";
		if (am & FAM_READ) mode="wb+";
	} else {
		if (am & FAM_READ) mode="rb";
		if (am & FAM_WRITE) mode="rb+";
	}
	if (mode) {
		file=fopen(filename, mode);
		if (!file) {
			set_error(errno | STERR_SYSTEM);
			if ((stream_error_func) && (stream_error_func(this)==SERR_RETRY))
				goto RETRY;
			return 0;
		}
	}
	return ht_streamfile::set_access_mode(am);
}

FILEOFS ht_file::tell()
{
	return offset;
}

int	ht_file::truncate(UINT newsize)
{
/* FIXME: there must be a better way... */
#define COPYBUFSIZE	32*1024
	char tempfile[PATH_MAX+20];
	sys_dirname(filename, tempfile);

	strcat(tempfile, "/");

	char *n=tempfile+strlen(tempfile);
	int i=0;
	FILE *exists;
	do {
		sprintf(n, "ht%06d.tmp", i++);
		errno=0;
		exists=fopen(tempfile, "rb");
		if (exists) fclose(exists);
	} while ((errno!=ENOENT) || exists);
	
	int e=0;

	int old_access_mode=access_mode;
	set_access_mode(FAM_UNDEFINED);

	e=rename(filename, tempfile);
	if (e==0) {
		FILE *o=fopen(tempfile, "rb");
		FILE *n=fopen(filename, "wb");
		void *buf=malloc(COPYBUFSIZE);
		while (newsize) {
			UINT q=fread(buf, 1, COPYBUFSIZE, o);
			if (!q) {
				e=EIO;
				set_error(e | STERR_SYSTEM);
				break;
			}
			if (q>newsize) q=newsize;
			fwrite(buf, 1, q, n);
			newsize-=q;
		}
		fclose(o);
		fclose(n);
		remove(tempfile);
		free(buf);
	} else {
		set_error(e | STERR_SYSTEM);
	}
	set_access_mode(old_access_mode);
	
	return e;
}

UINT	ht_file::write(void *buf, UINT size)
{
	UINT r=fwrite(buf, 1, size, file);
	offset+=r;
	return r;
}

/*
 *	CLASS ht_memmap_file
 */

void ht_memmap_file::init(byte *b)
{
	ht_streamfile::init();
	buf=b;
	pos=0;
}

void ht_memmap_file::done()
{
}

char *ht_memmap_file::get_name()
{
	return "memmap";
}

UINT ht_memmap_file::get_size()
{
	return 0;	/* not implemented */
}

UINT ht_memmap_file::read(void *b, UINT size)
{
	memmove(b, (char*)buf+pos, size);
	pos+=size;
	return size;
}

int ht_memmap_file::seek(FILEOFS offset)
{
	pos=offset;
	return 1;
}

FILEOFS ht_memmap_file::tell()
{
	return pos;
}

UINT ht_memmap_file::write(void *b, UINT size)
{
	memmove(((byte*)buf)+pos, b, size);
	pos+=size;
	return size;
}

/*
 *	CLASS ht_mem_file
 */

#define HTMEMFILE_INITIAL_SIZE		1024
#define HTMEMFILE_GROW_FACTOR_NOM	4
#define HTMEMFILE_GROW_FACTOR_DENOM	3

void ht_mem_file::init()
{
	ht_mem_file::init(0, HTMEMFILE_INITIAL_SIZE);
}

void ht_mem_file::init(FILEOFS o, UINT size)
{
	ht_streamfile::init();
	ofs=o;
	buf=(byte*)malloc(size);
	bufsize=size;
	memset(buf, 0, size);
	pos=0;
	dsize=0;
}

void ht_mem_file::done()
{
	free(buf);
	ht_stream::done();
}

void *ht_mem_file::bufptr()
{
	return buf;
}

void ht_mem_file::extendbuf()
{
	UINT oldbufsize=bufsize;
	bufsize=bufsize * HTMEMFILE_GROW_FACTOR_NOM / HTMEMFILE_GROW_FACTOR_DENOM;
	byte *t=(byte*)malloc(bufsize);
	memset(t, 0, bufsize);
	memmove(t, buf, oldbufsize);
	free(buf);
	buf=t;
}

UINT ht_mem_file::get_size()
{
	return dsize;
}

char *ht_mem_file::get_name()
{
	return "memstream";
}

UINT ht_mem_file::read(void *buf, UINT size)
{
	if (pos+size>bufsize) {
		memset(buf, 0, size);
		if (bufsize-pos+1>0) memmove(buf, this->buf+pos, bufsize-pos+1);
		size=bufsize-pos;
	} else {
		memmove(buf, (char*)this->buf+pos, size);
	}
	pos+=size;
	return size;
}

int ht_mem_file::seek(FILEOFS o)
{
	if (o<ofs) return 0;
	pos=o-ofs;
	return 1;
}

FILEOFS ht_mem_file::tell()
{
	return pos+ofs;
}

UINT ht_mem_file::write(void *b, UINT size)
{
	while (pos+size>=bufsize) extendbuf();
	memmove(((byte*)buf)+pos, b, size);
	pos+=size;
	if (pos>dsize) dsize=pos;
	return size;
}

/*
 *	string stream functions
 */

// FIXME: more dynamical solution appreciated
#define REASONABLE_STRING_LIMIT	256
 
char *fgetstrz(ht_streamfile *file)
{
	FILEOFS o=file->tell();
/* get string size */
	char buf[64];
	int s, z=0;
	int found=0;
	while (!found) {
		if (!(s=file->read(buf, 64))) break;
		for (int i=0; i<s; i++) {
			z++;
			if (!buf[i]) {
				found=1;
				break;
			}
		}
	}
	if (!z) return 0;
/* read string */
	if (z>REASONABLE_STRING_LIMIT) z=REASONABLE_STRING_LIMIT;
	char *str=(char *)malloc(z);
	file->seek(o);
	file->read(str, z);
	str[z-1]=0;
	return str;
}

char *getstrz(ht_stream *stream)
{
/* get string size */
	char buf[REASONABLE_STRING_LIMIT];
	int s, z=0;
	while (1) {
		if (!(s=stream->read(buf+z, 1))) break;
			 z++;
		if (z>=REASONABLE_STRING_LIMIT) {
				z=REASONABLE_STRING_LIMIT;
				    break;
		}
		if (!buf[z-1]) break;
	}
	if (!z) return 0;
	char *str=(char *)malloc(z);
	memmove(str, buf, z-1);
	str[z-1]=0;
	return str;
}

void putstrz(ht_stream *stream, char *str)
{
	stream->write(str, strlen(str)+1);
}

char *getstrp(ht_stream *stream)
{
	unsigned char l;
	stream->read(&l, 1);
	char *str=(char*)malloc(l+1);
	stream->read(str, l);
	*(str+l)=0;
	return str;
}

void putstrp(ht_stream *stream, char *str)
{
	unsigned char l=strlen(str);
	stream->write(&l, 1);
	stream->write(str, l);
}

char *getstrw(ht_stream *stream)
{
	short t;
	int l=0;
	stream->read(&l, 2);
	char	*a=(char*)malloc(l+1);
	for (int i=0; i<l; i++) {
		stream->read(&t, 2);
		*(a+i)=(char)t;
	}
	*(a+l)=0;
	return a;
}

void putstrw(ht_stream *stream, char *str)
{
	/* FIXME: someone implement me ? */
}

